/*********************************************************************
 * Reboot/Program Killer Joystick Daemon (jsr.c)
 * 
 * (c) 1994,1998  Scott Heavner (sdh@po.cwru.edu)
 *           Version 1.2 - July 1, 1998
 *********************************************************************
 * 
 *    jsr is a daemon that will perform certain commands when it 
 * detects joystick button action.  The actions I have chosen to
 * watch for are both buttons down and the release of one or
 * both buttons after both have been detected.
 *
 *    As it is set up now, if you press and release both buttons,
 * the computer will reboot.  If you press both buttons, release
 * one, but hold the other for longer than a second, then release
 * it, it will execute a "killall X" which should kill all active
 * X sessions, which is quite useful if the X server hangs.
 * You are welcome to plug in your own commands (You might have to
 * if your X process doesn't show up as "X :0").  The defaults
 * can be adjusted using ./configure (see INSTALL).
 * 
 *    All the defaults can also be set with command line arguments.
 * jsr -r "my new reboot command"            (or --reboot-cmd)
 * jsr -a "my new alt command"               (or --alt-cmd)
 * jsr -t my_new_button_timeout_in_seconds   (or --button-timeout)
 * jsr -d path_to_joystick_device            (or --device)
 *
 *    The commands are executed via a system() call and can be
 * any command you would issue to sh on the command line (script files,
 * executable files, sh builtins) with any arguments you like.
 *********************************************************************
 *    If you are using version 1.2.7 or newer of the joystick module,
 * jsr will use select() to read the joystick device.  This cuts down
 * on overhead as compared the polling version, which does a read
 * on the joystick device every 10ms.  If you upgrade your joystick
 * driver, from version < 1.2.7, you must recompile jsr to take
 * advantage of select().
 *********************************************************************
 *    The joystick should probably be looked upon as a temporary
 * solution.  I'm not using my joystick for anything and 
 * didn't feel like buiding a switch or circuit to simulate a UPS
 * shutdown request on the serial port.
 *
 *    Currently, the program keeps the joystick device open all the 
 * time and other programs cannot use it.  It might be nicer to check
 * less frequently and open/close the device, waiting for the
 * completion of any process which might have opened it.  For my use,
 * it is safer this way, I'm not using the joystick for anything else,
 * so I'd rather have my reboot available at all times.
 *
 *    Someone might say why stop at two button combinations -- because
 * two is enough and I don't want to have to remember even that many. 
 * My current choice is orientation independent and pretty easy to 
 * remember.  I also thought about a cancel option, but I don't really
 * need that either.
 *********************************************************************
 * You can rebuild the binary and install it in /sbin
 *
 *   ./configure ; make clean ; make ; make install
 *
 * If you don't want to bother with configure, compile with
 *    gcc -O2 -s -DSTDC_HEADERS=1 -DHAVE_FCNTL_H=1 \
                 -DHAVE_SYS_TIME_H=1 -DHAVE_SYS_TYPES_H=1 \
                 -DHAVE_UNISTD_H=1 -DHAVE_LINUX_JOYSTICK_H=1\
                 -DTIME_WITH_SYS_TIME=1 -DHAVE_SELECT=1\
                 -DHAVE_SYS_IOCTL jsr.c -o jsr
 *
 * Test:         ./jsr -r "echo Reboot" -a "echo Alt" --debug
 *
 * Add to rc:    You need to start jsr every time your computer is
 *               restarted.  Somehow you have to add an entry into
 *               your rc heirarchy, you may need to modify /etc/rc,
 *               /etc/init.d/boot, or /etc/init.d/<yourscript>?
 *               (what's the standard these days??)
 *********************************************************************
 * Requires:     joystick-0.7 module or newer (from sunsite)
 *
 *               Latest joystick module should be at: 
 *               http://atrey.karlin.mff.cuni.cz/~vojtech/joystick/
 *********************************************************************
 * jsr updates _may_ be available at
 *               http://www.geocities.com/CapeCanaveral/Lab/7731/jsr.html
 *********************************************************************
 * Version 1.0 - December 5, 1994
 * ------------------------------
 * - initial public release
 *
 * Version 1.1 - June 30, 1998
 * ---------------------------
 * - update for new joystick driver 1.2.7 allowing select() call
 * - change default commands to "/sbin/reboot" and "killall X", using
 *   wrong path to reboot and grep script on ps output
 * - add --help and --version info
 * - add auto fork() and --debug option to override fork()
 * - add configure
 * - change dist. policy to GPL
 *
 * Version 1.2 - July 1, 1998
 * ----------------------------
 * - add runtime checks for module version, without them may try to
 *   use select() with old version of driver which causes jsr to
 *   eat as much CPU as possible
 * - add a bunch of debugging printf's enabled using 
 *   ./configure --enable-debug
 *********************************************************************/

#ifndef DEFAULT_DEVICE
#define DEFAULT_DEVICE    "/dev/js0" /* Which joystick to use (usually js0 or js1) */
#endif

/* ./configure will search for reboot command or override with ./configure --with-reboot="" */
#ifndef REBOOT_CMD
#define REBOOT_CMD        "/sbin/reboot"
#endif

/* Choose your alt_cmd below or using ./configure --with-altcmd="" */
#ifndef ALT_CMD
#define ALT_CMD           "killall X" /* Kill all running X sessions*/
/* #define ALT_CMD        "chvt 1"    /* Change to a sane VT */
/* #define ALT_CMD        "textmode"  /* Reset after a SVGALIB crash */
#endif

#define DEFAULT_TIMEOUT   1        /* Timeout (sec) for altcmd selection 
                                    *  set 0 for always or huge for never */

#define JSR_VERSION        0x010200

#if HAVE_LINUX_JOYSTICK_H
#include <linux/joystick.h>
#else
#error *** You need to install the linux joystick module and header files ***
#endif

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if STDC_HEADERS
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>
#endif

/* Only needed with select() version of jsr */
#if HAVE_SELECT
#if TIME_WITH_SYS_TIME && HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#endif /* HAVE_SELECT */

/* Debugging printf() macros */
#if JSR_DEBUG
#define jsr_dmesg(a) fprintf(stderr,(a))
#define jsr_dversion(jsdriver_version)  fprintf(stderr, "Returned version %d.%d.%d\n", \
                                                (jsdriver_version) >> 16 & 0xff, \
                                                (jsdriver_version) >> 8 & 0xff, \
                                                (jsdriver_version) & 0xff)
#else
#define jsr_dmesg(a)
#define jsr_dversion(a)
#endif

/* Open the joystick device, exit on error */
int jsr_open_js(char *devicename)
{
        int fd;
        
        /* open device file */
        fd = open (devicename, O_RDONLY);
        if (fd < 0) {
                perror ("js");
                exit (1);
        }
        
        return fd;
}

/* Read the joystick device, exit on error */
void jsr_read_js(int fd, struct JS_DATA_TYPE *js)
{
        if( read (fd, js, JS_RETURN) != JS_RETURN) {
                perror ("js");
                exit (1);
        }
}

/* If joystick driver supports the select() call, use it */
void jsr_delay(int fd)
{
        static int poll_driver = -1;

#if defined JS_VERSION && (JS_VERSION >= 0x00010207) && HAVE_SELECT && HAVE_SYS_IOCTL_H
        fd_set select_set;
        struct timeval timeout;

        /* Now check the version of the module that is actually in memory */
        if (poll_driver < 0) {
                __u32 jsdriver_version;
                int result;
		jsr_dmesg("Checking version of currently loaded joystick module\n");
                if (!ioctl(fd, JSIOCGVERSION, &jsdriver_version)) {
		        jsr_dversion(jsdriver_version);
                        if (jsdriver_version >= 0x00010207) {
                                poll_driver = 0;
                                jsr_dmesg("Version > 1.2.7 - will use select()\n");
                        }
                } else {
                       jsr_dmesg("Can't read version from joystick module, assuming < 1.2.7\n");
                }
        }
        
        if (!poll_driver) {
                /* Use select() to idle the process while we wait for an action */
                FD_ZERO(&select_set);		/* Have to reset each time we get an event ? */
                FD_SET(fd, &select_set);
                timeout.tv_sec =  32767;	/* What is the standard for a long time?? */
                timeout.tv_usec = 0; 		/*   - How about 1/yr we fall out of the select() */
                jsr_dmesg("waiting in select . . .\n");
                select(FD_SETSIZE, &select_set, 0, 0, &timeout);
        }
#endif
#if JSR_DEBUG
	if (poll_driver==-1) {
                jsr_dmesg("Using polling driver . . .\n");
                poll_driver = 1;
        }
#endif
        /* give other processes a chance - we should poll @ 10Hz maybe??  Start with 100Hz
         * It's not like we are really doing anything to load the system */
        if (poll_driver)
                usleep(10000UL);
}

/* Make some noise */
void jsr_beep()
{
        fprintf(stderr,"\007");
        fflush(stderr);
}

/* Do the command */
void jsr_exec(char *cmd)
{
        if (!cmd)
                return;
        jsr_beep();
#       if JSR_DEBUG
                fprintf(stderr,"Executing system command: \"%s\"\n",cmd);
#       endif
        system(cmd);
        usleep(10000UL);  /* Debouncing for the select version */
}

/* Display version info */
void jsr_version()
{
        printf("jsr %d.%d.%d\n", 
               JSR_VERSION >> 16 & 0xff,
               JSR_VERSION >> 8 & 0xff,
               JSR_VERSION & 0xff );
}

/* Display help and version */
void jsr_help()
{
    jsr_version();
    printf("-r (--reboot-cmd)      System command to exec on reboot signal\n"
           "                       Default=\"%s\"\n",REBOOT_CMD);
    printf("-a (--alt-cmd)         System command to exec on alternate signal\n"
           "                       Default=\"%s\"\n",ALT_CMD);
    printf("-d (--device)          Full path to joystice device "
                                  "(default=\"%s\")\n",DEFAULT_DEVICE);
    printf("-t (--button-timeout)  Time (s) used to distinguish between reboot/alt\n"
           "                       Default=\"%d\"\n",DEFAULT_TIMEOUT);
    printf("-h (--help)            Display this help\n");
    printf("-v (--version)         Display version info\n");
    printf("--debug                Don't fork() and background jsr\n");
    printf("\nWhen the daemon notices that both buttons are pressed, it then waits\n"
	   "for the release of one or both buttons. If both buttons are released\n"
	   "simultaneously, jsr will exec the reboot command.  If one button is\n"
	   "released and the other is released <button-timeout> seconds or later,\n"
	   "the alternate command is executed.\n");
}

int main(int argc, char **argv)
{
        int    fd, done=0, dontfork=0;
        struct JS_DATA_TYPE js;
        time_t time1, button_timeout = DEFAULT_TIMEOUT;
        char   *reboot_cmd = REBOOT_CMD;
        char   *alt_cmd    = ALT_CMD;
        char   *devicename = DEFAULT_DEVICE;
        struct option long_options[] = {
                {"reboot-cmd", 1, 0, 'r'},
                {"alt-cmd", 1, 0, 'a'},
                {"device", 1, 0, 'd'},
                {"button-timeout", 1, 0, 't'},
                {"help", 0, 0, 'h'},
                {"version", 0, 0, 'v'},
                {"debug", 0, 0, 'D'},
                {0,0,0,0}
        };
        
        /* Parse the command line options */
        while(!done) {
                switch(getopt_long (argc, argv, "r:a:d:t:hvD", long_options, &fd)) {
                case -1:
                        done = 1;
                        break;
                case 'D':
                        dontfork = 1;
                        break;
                case 'r':
                        reboot_cmd = optarg;
                        break;
                case 'a':
                        alt_cmd = optarg;
                        break;
                case 'd':
                        devicename = optarg;
                        break;
                case 't':
                        sscanf(optarg,"%ld",&button_timeout);
                        break;
                case 'v':
                        jsr_version();
                        exit(0);
                case 'h':
                        jsr_help();
                        exit(0);
                }
        }
        
#if !JSR_DEBUG
        /* go to background */
        if (!dontfork) {
                switch(fork()) {
                        case -1: perror("fork()"); /* error */
                                exit(-1);
                        case  0: break;          /* child */
                        default: exit(0);        /* parent */
                }
        }
#else
	jsr_dmesg("Compile time debugging enabled - skipping fork().\n");
#endif
    
        /* open device file */
        fd = jsr_open_js(devicename);

        while (1) {
                /* Check the joystick device */
                jsr_read_js(fd,&js); 
                
                /* Check for both buttons down, wait for at least one to release */
                if ((js.buttons&3)==3) {

                        jsr_dmesg("Noticed both buttons down\n");

                        /* Output a beep */
                        jsr_beep();
                        
                        /* Wait for at least one release */
                        while( (js.buttons&3)==3)
                                jsr_read_js(fd,&js); 

                        jsr_dmesg("Noticed at least one was released\n");
                        
                        /* Grab the current time */ 
                        time1 = time(NULL);
                        
                        /* Wait for the other button to be released */
                        while( (js.buttons&3))
                                jsr_read_js(fd,&js);

                        jsr_dmesg("Noticed both were released\n");
                        
                        /* If the single button was held more than one second,
                         *      exec the alternate command, else reboot */ 
                        if ((time(NULL)-time1)>=button_timeout) {
                                jsr_exec(alt_cmd);
                        } else {
                                jsr_exec(reboot_cmd);
                        }
                }
                
                jsr_delay(fd);
        }
        
        return 0; 
}


